/*
 * SIDH: an efficient supersingular isogeny cryptography library
 * x64 assembly code for P503 using PMNS
 */

.intel_syntax noprefix

/* Format function and variable names for Mac OS X */
#if defined(__APPLE__)
  #define fmt(f) _##f
#else
  #define fmt(f) f
#endif

/* Registers that are used for parameter passing: */
#define reg_p1 rdi
#define reg_p2 rsi
#define reg_p3 rdx

#include "pmns_internal.h"

.text
/*
 * Negation in GF(p) in PMNS
 * Operation: a [reg_p1] = -a [reg_p1]
 */
////.global fmt(fpneg503)
fmt(fpneg503):
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx
  push  rbp

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+8]
  mov   r10, [reg_p1+16]
  mov   r11, [reg_p1+24]
  mov   r12, [reg_p1+32]
  mov   r13, [reg_p1+40]
  mov   r14, [reg_p1+48]
  mov   r15, [reg_p1+56]
  mov   rbp, [reg_p1+64]
  mov   rbx, [reg_p1+72]

  mov   rcx, PMNS_GAMMA
  mov   rdx, PMNS_SUB_CST0
  mov   rsi, PMNS_K_MINUS_EP1

  sub   rdx, r8
  cmp   rdx, rsi
  sbb   rax, rax
  and   rax, rcx
  sub   rdx, rax

  mov   rsi, PMNS_SUB_CST1

  mov   r8, rsi
  sub   r8, r9
  neg   rax
  adc   r8, 0
  bt    r8, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r8, rax

  mov   r9, rsi
  sub   r9, r10
  neg   rax
  adc   r9, 0
  bt    r9, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r9, rax

  mov   r10, rsi
  sub   r10, r11
  neg   rax
  adc   r10, 0
  bt    r10, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r10, rax

  mov   r11, rsi
  sub   r11, r12
  neg   rax
  adc   r11, 0
  bt    r11, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r11, rax

  mov   r12, rsi
  sub   r12, r13
  neg   rax
  adc   r12, 0
  bt    r12, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r12, rax

  mov   r13, rsi
  sub   r13, r14
  neg   rax
  adc   r13, 0
  bt    r13, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r13, rax

  mov   r14, rsi
  sub   r14, r15
  neg   rax
  adc   r14, 0
  bt    r14, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r14, rax

  mov   r15, rsi
  sub   r15, rbp
  neg   rax
  adc   r15, 0
  bt    r15, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   r15, rax

  mov   rbp, rsi
  sub   rbp, rbx
  neg   rax
  adc   rbp, 0
  bt    rbp, PMNS_K
  sbb   rax, rax
  and   rax, rcx
  sub   rbp, rax

  neg   rax
  sbb   rax, rax
  and   rax, 3
  add   rdx, rax

  mov   [reg_p1], rdx
  mov   [reg_p1+8], r8
  mov   [reg_p1+16], r9
  mov   [reg_p1+24], r10
  mov   [reg_p1+32], r11
  mov   [reg_p1+40], r12
  mov   [reg_p1+48], r13
  mov   [reg_p1+56], r14
  mov   [reg_p1+64], r15
  mov   [reg_p1+72], rbp

  pop   rbp
  pop   rbx
  pop   r15
  pop   r14
  pop   r13
  pop   r12
  ret


/*
 * Double subtraction in PMNS without coefficients reduction
 * Operation: c [reg_p3] = c [reg_p3] - a [reg_p1] - b [reg_p2]
 */
.global fmt(mp_dblsub503x2_pmns)
fmt(mp_dblsub503x2_pmns):
  push  r12
  push  r13
  push  r14
  push  r15

  mov   r8, [reg_p3]
  mov   r9, [reg_p3+8]
  mov   r10, [reg_p3+16]
  mov   r11, [reg_p3+24]
  mov   r12, [reg_p3+32]
  mov   r13, [reg_p3+40]
  mov   r14, [reg_p3+48]
  mov   r15, [reg_p3+56]

  sub   r8, [reg_p1]
  sbb   r9, [reg_p1+8]

  sub   r10, [reg_p1+16]
  sbb   r11, [reg_p1+24]

  sub   r12, [reg_p1+32]
  sbb   r13, [reg_p1+40]

  sub   r14, [reg_p1+48]
  sbb   r15, [reg_p1+56]

  sub   r8, [reg_p2]
  sbb   r9, [reg_p2+8]

  sub   r10, [reg_p2+16]
  sbb   r11, [reg_p2+24]

  sub   r12, [reg_p2+32]
  sbb   r13, [reg_p2+40]

  sub   r14, [reg_p2+48]
  sbb   r15, [reg_p2+56]

  mov   [reg_p3], r8
  mov   [reg_p3+8], r9
  mov   [reg_p3+16], r10
  mov   [reg_p3+24], r11
  mov   [reg_p3+32], r12
  mov   [reg_p3+40], r13
  mov   [reg_p3+48], r14
  mov   [reg_p3+56], r15

  mov   r8, [reg_p3+64]
  mov   r9, [reg_p3+72]
  mov   r10, [reg_p3+80]
  mov   r11, [reg_p3+88]
  mov   r12, [reg_p3+96]
  mov   r13, [reg_p3+104]
  mov   r14, [reg_p3+112]
  mov   r15, [reg_p3+120]

  sub   r8, [reg_p1+64]
  sbb   r9, [reg_p1+72]

  sub   r10, [reg_p1+80]
  sbb   r11, [reg_p1+88]

  sub   r12, [reg_p1+96]
  sbb   r13, [reg_p1+104]

  sub   r14, [reg_p1+112]
  sbb   r15, [reg_p1+120]

  sub   r8, [reg_p2+64]
  sbb   r9, [reg_p2+72]

  sub   r10, [reg_p2+80]
  sbb   r11, [reg_p2+88]

  sub   r12, [reg_p2+96]
  sbb   r13, [reg_p2+104]

  sub   r14, [reg_p2+112]
  sbb   r15, [reg_p2+120]

  mov   [reg_p3+64], r8
  mov   [reg_p3+72], r9
  mov   [reg_p3+80], r10
  mov   [reg_p3+88], r11
  mov   [reg_p3+96], r12
  mov   [reg_p3+104], r13
  mov   [reg_p3+112], r14
  mov   [reg_p3+120], r15

  mov   r8, [reg_p3+128]
  mov   r9, [reg_p3+136]
  mov   r10, [reg_p3+144]
  mov   r11, [reg_p3+152]

  sub   r8, [reg_p1+128]
  sbb   r9, [reg_p1+136]

  sub   r10, [reg_p1+144]
  sbb   r11, [reg_p1+152]

  sub   r8, [reg_p2+128]
  sbb   r9, [reg_p2+136]

  sub   r10, [reg_p2+144]
  sbb   r11, [reg_p2+152]

  mov   [reg_p3+128], r8
  mov   [reg_p3+136], r9
  mov   [reg_p3+144], r10
  mov   [reg_p3+152], r11

  pop   r15
  pop   r14
  pop   r13
  pop   r12
  ret

/*
 * Substraction in PMNS when coeffs are on two words
 * Operation: c [reg_p2] = a [reg_p1] - b [reg_p2]
 */
.global fmt(mp_sub503x2_pmns)
fmt(mp_sub503x2_pmns):
  mov   r8, [reg_p1]
  mov   r9, [reg_p1+0x8]
  mov   r10, [reg_p1+0x10]
  mov   r11, [reg_p1+0x18]

  mov   rax, PMNS_SUBx2_CST0_0
  mov   rcx, PMNS_SUBx2_CST0_1

  add   r8, rax
  mov   rax, PMNS_SUBx2_CST1_0
  adc   r9, rcx
  mov   rcx, PMNS_SUBx2_CST1_1

  sub   r8, [reg_p2]
  sbb   r9, [reg_p2+0x8]

  add   r10, rax
  adc   r11, rcx

  sub   r10, [reg_p2+0x10]
  sbb   r11, [reg_p2+0x18]

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10
  mov   [reg_p3+0x18], r11

  mov   r8, [reg_p1+0x20]
  mov   r9, [reg_p1+0x28]
  mov   r10, [reg_p1+0x30]
  mov   r11, [reg_p1+0x38]

  add   r8, rax
  adc   r9, rcx
  add   r10, rax
  adc   r11, rcx

  sub   r8, [reg_p2+0x20]
  sbb   r9, [reg_p2+0x28]
  sub   r10, [reg_p2+0x30]
  sbb   r11, [reg_p2+0x38]

  mov   [reg_p3+0x20], r8
  mov   [reg_p3+0x28], r9
  mov   [reg_p3+0x30], r10
  mov   [reg_p3+0x38], r11

  mov   r8, [reg_p1+0x40]
  mov   r9, [reg_p1+0x48]
  mov   r10, [reg_p1+0x50]
  mov   r11, [reg_p1+0x58]

  add   r8, rax
  adc   r9, rcx
  add   r10, rax
  adc   r11, rcx

  sub   r8, [reg_p2+0x40]
  sbb   r9, [reg_p2+0x48]
  sub   r10, [reg_p2+0x50]
  sbb   r11, [reg_p2+0x58]

  mov   [reg_p3+0x40], r8
  mov   [reg_p3+0x48], r9
  mov   [reg_p3+0x50], r10
  mov   [reg_p3+0x58], r11

  mov   r8, [reg_p1+0x60]
  mov   r9, [reg_p1+0x68]
  mov   r10, [reg_p1+0x70]
  mov   r11, [reg_p1+0x78]

  add   r8, rax
  adc   r9, rcx
  add   r10, rax
  adc   r11, rcx

  sub   r8, [reg_p2+0x60]
  sbb   r9, [reg_p2+0x68]
  sub   r10, [reg_p2+0x70]
  sbb   r11, [reg_p2+0x78]

  mov   [reg_p3+0x60], r8
  mov   [reg_p3+0x68], r9
  mov   [reg_p3+0x70], r10
  mov   [reg_p3+0x78], r11

  mov   r8, [reg_p1+0x80]
  mov   r9, [reg_p1+0x88]
  mov   r10, [reg_p1+0x90]
  mov   r11, [reg_p1+0x98]

  add   r8, rax
  adc   r9, rcx
  add   r10, rax
  adc   r11, rcx

  sub   r8, [reg_p2+0x80]
  sbb   r9, [reg_p2+0x88]
  sub   r10, [reg_p2+0x90]
  sbb   r11, [reg_p2+0x98]

  mov   [reg_p3+0x80], r8
  mov   [reg_p3+0x88], r9
  mov   [reg_p3+0x90], r10
  mov   [reg_p3+0x98], r11

  ret

/*
 * Montgomery reduction in PMNS
 * Operation: c [reg_p2] = redc(a [reg_p1])
 */
.global fmt(rdc_mont)
fmt(rdc_mont):
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx
  push  rbp

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+16]
  mov   rcx, [reg_p1+144]

  mov   rax, PMNS_GAMMA_OVER_E
  mul   r8
  add   rcx, rax
  mov   rax, PMNS_GAMMA2_OVER_E_MOD
  mul   r9
  add   rcx, rax

  mov   rdx, PMNS_GAMMA
  mov   rbx, [reg_p1+80]
  mov   rbp, [reg_p1+88]
  mov   r14, [reg_p1+96]
  mov   r15, [reg_p1+104]
  mov   r12, [reg_p1+112]
  mov   r13, [reg_p1+120]
  mov   r10, [reg_p1+128]
  mov   r11, [reg_p1+136]

  mulx  rax, rcx, rcx
  add   rcx, r10
  adc   r11, rax

  mulx  rax, rcx, rcx
  add   rcx, r12
  adc   r13, rax

  mulx  rax, rcx, rcx
  add   rcx, r14
  adc   r15, rax

  mulx  rax, rcx, rcx
  add   rcx, rbx
  adc   rbp, rax

  mov   [reg_p2+40], rbp
  mov   [reg_p2+48], r15
  mov   [reg_p2+56], r13
  mov   [reg_p2+64], r11

  mov   rbx, [reg_p1+8]
  mov   rbp, [reg_p1+24]
  mov   r14, [reg_p1+32]
  mov   r15, [reg_p1+40]
  mov   r12, [reg_p1+48]
  mov   r13, [reg_p1+56]
  mov   r10, [reg_p1+64]
  mov   r11, [reg_p1+72]

  mulx  rax, rcx, rcx
  add   rcx, r10
  adc   r11, rax

  mulx  rax, rcx, rcx
  add   rcx, r12
  adc   r13, rax

  mulx  rax, rcx, rcx
  add   rcx, r14
  adc   r15, rax

  mulx  rax, rcx, rcx
  add   rcx, r9
  adc   rbp, rax

  mulx  rax, rcx, rcx
  add   rcx, r8
  adc   rbx, rax

  mov   [reg_p2], rbx
  mov   [reg_p2+8], rbp
  mov   [reg_p2+16], r15
  mov   [reg_p2+24], r13
  mov   [reg_p2+32], r11

  mov   r8, [reg_p1+144]
  mov   r9, [reg_p1+152]
  mov   rdx, PMNS_GAMMA_OVER_E

  mulx  rax, rcx, rcx
  add   rcx, r8
  adc   r9, rax

  mov   [reg_p2+72], r9

  pop   rbp
  pop   rbx
  pop   r15
  pop   r14
  pop   r13
  pop   r12
  ret

/*
 * Schoolbook multiplication for polynomial of degree 4
 * rdx is reserved for internal usage, registers parameters must not be rdx.
 */
.macro MUL_DEG4_SCHOOL M0, M1, C, T0, T1, T2, T3, T4, T5, T6, T7, T8, T9
    mov    rdx, \M0
    mulx   \T1, \T0, \M1
    mulx   \T3, \T2, 8\M1
    mov    \C, \T0
    mov    8\C, \T1   /* C0 = \M0*\M1 */
    mulx   \T5, \T4, 16\M1
    mulx   \T7, \T6, 24\M1
    mulx   \T9, \T8, 32\M1

    mov    rdx, 8\M0
    mulx   \T1, \T0, \M1
    add    \T2, \T0
    adc    \T3, \T1
    mov    16\C, \T2
    mov    24\C, \T3  /* C1 = \M0*8\M1 + 8\M0*\M1 */
    mulx   \T1, \T0, 8\M1
    add    \T4, \T0
    adc    \T5, \T1
    mulx   \T3, \T2, 16\M1
    add    \T6, \T2
    adc    \T7, \T3
    mulx   \T1, \T0, 24\M1
    add    \T8, \T0
    adc    \T9, \T1
    mulx   \T3, \T2, 32\M1

    mov    rdx, 16\M0
    mulx   \T1, \T0, \M1
    add    \T4, \T0
    adc    \T5, \T1
    mov    32\C, \T4
    mov    40\C, \T5  /* C2 = \M0*16\M1 + 8\M0*8\M1 + 16\M0*\M1 */
    mulx   \T1, \T0, 8\M1
    add    \T6, \T0
    adc    \T7, \T1
    mulx   \T5, \T4, 16\M1
    add    \T8, \T4
    adc    \T9, \T5
    mulx   \T1, \T0, 24\M1
    add    \T2, \T0
    adc    \T3, \T1
    mulx   \T5, \T4, 32\M1

    mov    rdx, 24\M0
    mulx   \T1, \T0, \M1
    add    \T6, \T0
    adc    \T7, \T1
    mov    48\C, \T6
    mov    56\C, \T7  /* C3 = \M0*24\M1 + 8\M0*16\M1 + 16\M0*8\M1 + 24\M0*\M1 */
    mulx   \T1, \T0, 8\M1
    add    \T8, \T0
    adc    \T9, \T1
    mulx   \T7, \T6, 16\M1
    add    \T2, \T6
    adc    \T3, \T7
    mulx   \T1, \T0, 24\M1
    add    \T4, \T0
    adc    \T5, \T1
    mulx   \T7, \T6, 32\M1

    mov    rdx, 32\M0
    mulx   \T1, \T0, \M1
    add    \T8, \T0
    adc    \T9, \T1
    mov    64\C, \T8
    mov    72\C, \T9  /* C4 = \M0*32\M1 + 8\M0*24\M1 + ... + 32\M0*\M1 */
    mulx   \T1, \T0, 8\M1
    add    \T2, \T0
    adc    \T3, \T1
    mulx   \T9, \T8, 16\M1
    add    \T4, \T8
    adc    \T5, \T9
    mulx   \T1, \T0, 24\M1
    add    \T6, \T0
    adc    \T7, \T1
    mulx   \T9, \T8, 32\M1

    mov    80\C, \T2
    mov    88\C, \T3 /* C5 = 8\M0*32\M1 +16\M0*24\M1 +24\M0*16\M1 +32\M0*8\M1 */
    mov    96\C, \T4
    mov    104\C, \T5 /* C6 = 16\M0*32\M1 + 24\M0*24\M1 + 32\M0*16\M1 */
    mov    112\C, \T6
    mov    120\C, \T7 /* C7 = 24\M0*32\M1 + 32\M0*24\M1 */
    mov    128\C, \T8
    mov    136\C, \T9 /* C8 = 32\M0*32\M1 */
.endm

/*
 * ADK algorithm for multiplication of polynomial of degree 4
 * Using Eq. 2 of "Missing a trick: Karatsuba variations" by Michael Scott
 * [ url: https://eprint.iacr.org/2015/1247.pdf ]
 * rdx is reserved for internal usage, registers parameters must not be rdx.
 */
.macro MUL_DEG4_ADK M0, M1, C, T0, T1, T2, T3, T4, T5, T6, T7, T8, T9

  mov   rdx, \M0
  mulx  \T1, \T0, \M1
  mov   rdx, 8\M0
  mulx  \T9, \T8, 8\M1

  mov   \C, \T0
  mov   8\C, \T1    /* C0 = \M0*\M1 */

  add   \T0, \T8
  adc   \T1, \T9

  mov   rdx, \M0
  mov   \T3, \M1
  add   rdx, 8\M0
  add   \T3, 8\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  mov   16\C, \T2
  mov   24\C, \T3   /* C1 = \M0*8\M1 + 8\M0*\M1 */

  mov   rdx, 16\M0
  mulx  \T7, \T6, 16\M1

  add   \T0, \T6
  adc   \T1, \T7

  mov   rdx, \M0
  mov   \T3, \M1
  add   rdx, 16\M0
  add   \T3, 16\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  add   \T2, \T8
  adc   \T3, \T9
  add   \T2, \T8
  adc   \T3, \T9

  mov   32\C, \T2
  mov   40\C, \T3   /* C2 = \M0*16\M1 + 8\M0*8\M1 + 16\M0*\M1 */

  mov   rdx, 24\M0
  mulx  \T5, \T4, 24\M1

  add   \T0, \T4
  adc   \T1, \T5

  mov   rdx, 8\M0
  mov   \T3, 8\M1
  add   rdx, 16\M0
  add   \T3, 16\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  mov   96\C, \T4  /* temporary */
  mov   104\C, \T5 /* temporary */

  mov   rdx, \M0
  mov   \T5, \M1
  add   rdx, 24\M0
  add   \T5, 24\M1
  mulx  \T5, \T4, \T5

  add   \T2, \T4
  adc   \T3, \T5

  mov   48\C, \T2
  mov   56\C, \T3   /* C3 = \M0*24\M1 + 8\M0*16\M1 + 16\M0*8\M1 + 24\M0*\M1 */

  mov   rdx, 32\M0
  mulx  \T5, \T4, 32\M1

  add   \T0, \T4
  adc   \T1, \T5

  mov   128\C, \T4
  mov   136\C, \T5  /* C8 = 32\M0*32\M1 */

  mov   rdx, 8\M0
  mov   \T3, 8\M1
  add   rdx, 24\M0
  add   \T3, 24\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  mov   rdx, \M0
  mov   \T5, \M1
  add   rdx, 32\M0
  add   \T5, 32\M1
  mulx  \T5, \T4, \T5

  add   \T2, \T4
  adc   \T3, \T5

  add   \T2, \T6
  adc   \T3, \T7
  add   \T2, \T6
  adc   \T3, \T7

  mov   64\C, \T2
  mov   72\C, \T3   /* C4 = \M0*32\M1 + 8\M0*24\M1 + ... + 32\M0*\M1 */

  sub   \T0, \C
  sbb   \T1, 8\C

  mov   rdx, 16\M0
  mov   \T3, 16\M1
  add   rdx, 24\M0
  add   \T3, 24\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  mov   rdx, 8\M0
  mov   \T5, 8\M1
  add   rdx, 32\M0
  add   \T5, 32\M1
  mulx  \T5, \T4, \T5

  add   \T2, \T4
  adc   \T3, \T5

  mov   80\C, \T2
  mov   88\C, \T3   /* C5 = 8\M0*32\M1 +16\M0*24\M1 +24\M0*16\M1 +32\M0*8\M1 */

  sub   \T0, \T8
  sbb   \T1, \T9

  mov   \T8, 96\C
  mov   \T9, 104\C

  mov   rdx, 16\M0
  mov   \T3, 16\M1
  add   rdx, 32\M0
  add   \T3, 32\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  add   \T2, \T8
  adc   \T3, \T9
  add   \T2, \T8
  adc   \T3, \T9

  mov   96\C, \T2
  mov   104\C, \T3  /* C6 = 16\M0*32\M1 + 24\M0*24\M1 + 32\M0*16\M1 */

  sub   \T0, \T6
  sbb   \T1, \T7

  mov   rdx, 24\M0
  mov   \T3, 24\M1
  add   rdx, 32\M0
  add   \T3, 32\M1
  mulx  \T3, \T2, \T3

  sub   \T2, \T0
  sbb   \T3, \T1

  mov   112\C, \T2
  mov   120\C, \T3  /* C7 = 24\M0*32\M1 + 32\M0*24\M1 */
.endm


#define BASECASE_MACRO MUL_DEG4_SCHOOL
//#define BASECASE_MACRO MUL_DEG4_ADK
//#define SCHOOLBOOK_MUL 1

/*
 * Multiplication in GF(p) in PMNS (using one level of Karatsuba)
 * Operation: c [reg_p3] = a [reg_p1] * b [reg_p2]
 */
.global fmt(mp_mul)
fmt(mp_mul):
#ifndef SCHOOLBOOK_MUL
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx

  mov   rcx, reg_p3

  sub   rsp, 288

#if 0 // #ifdef __AVX2__
  vmovdqu ymm0, [reg_p1]
  vmovdqu ymm2, [reg_p2]
  mov   r8, [reg_p1+32]
  mov   r9, [reg_p2+32]

  vpaddq ymm1, ymm0, [reg_p1+40]
  vpaddq ymm3, ymm2, [reg_p2+40]
  add   r8, [reg_p1+72]
  add   r9, [reg_p2+72]

  vmovdqu [rsp], ymm1
  vmovdqu [rsp+40], ymm3
  mov   [rsp+32], r8
  mov   [rsp+72], r9
#else
  mov   r8, [reg_p1]
  mov   r9, [reg_p1+8]
  mov   r10, [reg_p1+16]
  mov   r11, [reg_p1+24]
  mov   r12, [reg_p1+32]
  add   r8, [reg_p1+40]
  add   r9, [reg_p1+48]
  add   r10, [reg_p1+56]
  add   r11, [reg_p1+64]
  add   r12, [reg_p1+72]
  mov   [rsp], r8
  mov   [rsp+8], r9
  mov   [rsp+16], r10
  mov   [rsp+24], r11
  mov   [rsp+32], r12

  mov   r8, [reg_p2]
  mov   r9, [reg_p2+8]
  mov   r10, [reg_p2+16]
  mov   r11, [reg_p2+24]
  mov   r12, [reg_p2+32]
  add   r8, [reg_p2+40]
  add   r9, [reg_p2+48]
  add   r10, [reg_p2+56]
  add   r11, [reg_p2+64]
  add   r12, [reg_p2+72]
  mov   [rsp+40], r8
  mov   [rsp+48], r9
  mov   [rsp+56], r10
  mov   [rsp+64], r11
  mov   [rsp+72], r12
#endif

  BASECASE_MACRO [rsp], [rsp+40], [rsp+144], r8, r9, r10, r11, r12, r13, r14, r15, rax, rbx

  BASECASE_MACRO [reg_p1], [reg_p2], [rcx], r8, r9, r10, r11, r12, r13, r14, r15, rax, rbx

  BASECASE_MACRO [reg_p1+40], [reg_p2+40], [rsp], r8, r9, r10, r11, r12, r13, r14, r15, rax, rbx


  mov   r8, [rcx]
  mov   r9, [rcx+8]
  mov   r10, [rcx+80]
  mov   r11, [rcx+88]
  mov   r12, [rsp+144]
  mov   r13, [rsp+152]
  mov   r14, [rsp+224]
  mov   r15, [rsp+232]
  mov   rax, [rsp]
  mov   rbx, [rsp+8]
  mov   rdi, [rsp+80]
  mov   rsi, [rsp+88]

  sub   r12, r8
  sbb   r13, r9
  sub   r14, r10
  sbb   r15, r11

  sub   r12, rax
  sbb   r13, rbx
  sub   r14, rdi
  sbb   r15, rsi

  add   r14, rax
  adc   r15, rbx

  add   r12, rdi
  adc   r13, rsi
  shld  rsi, rdi, 1
  shl   rdi
  add   r12, rdi
  adc   r13, rsi

  add   r8, r14
  adc   r9, r15
  shld  r15, r14, 1
  shl   r14
  add   r8, r14
  adc   r9, r15

  add   r12, r10
  adc   r13, r11

  mov   [rcx], r8
  mov   [rcx+8], r9
  mov   [rcx+80], r12
  mov   [rcx+88], r13


  mov   r8, [rcx+16]
  mov   r9, [rcx+24]
  mov   r10, [rcx+96]
  mov   r11, [rcx+104]
  mov   r12, [rsp+160]
  mov   r13, [rsp+168]
  mov   r14, [rsp+240]
  mov   r15, [rsp+248]
  mov   rax, [rsp+16]
  mov   rbx, [rsp+24]
  mov   rdi, [rsp+96]
  mov   rsi, [rsp+104]

  sub   r12, r8
  sbb   r13, r9
  sub   r14, r10
  sbb   r15, r11

  sub   r12, rax
  sbb   r13, rbx
  sub   r14, rdi
  sbb   r15, rsi

  add   r14, rax
  adc   r15, rbx

  add   r12, rdi
  adc   r13, rsi
  shld  rsi, rdi, 1
  shl   rdi
  add   r12, rdi
  adc   r13, rsi

  add   r8, r14
  adc   r9, r15
  shld  r15, r14, 1
  shl   r14
  add   r8, r14
  adc   r9, r15

  add   r12, r10
  adc   r13, r11

  mov   [rcx+16], r8
  mov   [rcx+24], r9
  mov   [rcx+96], r12
  mov   [rcx+104], r13


  mov   r8, [rcx+32]
  mov   r9, [rcx+40]
  mov   r10, [rcx+112]
  mov   r11, [rcx+120]
  mov   r12, [rsp+176]
  mov   r13, [rsp+184]
  mov   r14, [rsp+256]
  mov   r15, [rsp+264]
  mov   rax, [rsp+32]
  mov   rbx, [rsp+40]
  mov   rdi, [rsp+112]
  mov   rsi, [rsp+120]

  sub   r12, r8
  sbb   r13, r9
  sub   r14, r10
  sbb   r15, r11

  sub   r12, rax
  sbb   r13, rbx
  sub   r14, rdi
  sbb   r15, rsi

  add   r14, rax
  adc   r15, rbx

  add   r12, rdi
  adc   r13, rsi
  shld  rsi, rdi, 1
  shl   rdi
  add   r12, rdi
  adc   r13, rsi

  add   r8, r14
  adc   r9, r15
  shld  r15, r14, 1
  shl   r14
  add   r8, r14
  adc   r9, r15

  add   r12, r10
  adc   r13, r11

  mov   [rcx+32], r8
  mov   [rcx+40], r9
  mov   [rcx+112], r12
  mov   [rcx+120], r13


  mov   r8, [rcx+48]
  mov   r9, [rcx+56]
  mov   r10, [rcx+128]
  mov   r11, [rcx+136]
  mov   r12, [rsp+192]
  mov   r13, [rsp+200]
  mov   r14, [rsp+272]
  mov   r15, [rsp+280]
  mov   rax, [rsp+48]
  mov   rbx, [rsp+56]
  mov   rdi, [rsp+128]
  mov   rsi, [rsp+136]

  sub   r12, r8
  sbb   r13, r9
  sub   r14, r10
  sbb   r15, r11

  sub   r12, rax
  sbb   r13, rbx
  sub   r14, rdi
  sbb   r15, rsi

  add   r14, rax
  adc   r15, rbx

  add   r12, rdi
  adc   r13, rsi
  shld  rsi, rdi, 1
  shl   rdi
  add   r12, rdi
  adc   r13, rsi

  add   r8, r14
  adc   r9, r15
  shld  r15, r14, 1
  shl   r14
  add   r8, r14
  adc   r9, r15

  add   r12, r10
  adc   r13, r11

  mov   [rcx+48], r8
  mov   [rcx+56], r9
  mov   [rcx+128], r12
  mov   [rcx+136], r13


  mov   r8, [rcx+64]
  mov   r9, [rcx+72]
  mov   r10, [rsp+64]
  mov   r11, [rsp+72]
  mov   r12, [rsp+208]
  mov   r13, [rsp+216]

  sub   r12, r8
  sbb   r13, r9
  sub   r12, r10
  sbb   r13, r11

  add   r8, r10
  adc   r9, r11
  shld  r11, r10, 1
  shl   r10
  add   r8, r10
  adc   r9, r11

  mov   [rcx+64], r8
  mov   [rcx+72], r9
  mov   [rcx+144], r12
  mov   [rcx+152], r13


  add   rsp, 288

  pop   rbx
  pop   r15
  pop   r14
  pop   r13
  pop   r12
  ret
#else
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx

  mov   rcx, reg_p3

  mov   rdx, [reg_p1]
  mulx  r9, r8, [reg_p2]
  mulx  r11, r10, [reg_p2+0x8]
  mov   [rcx], r8
  mov   [rcx+0x8], r9   /* c0 contains a0*b0 */
  mulx  r13, r12, [reg_p2+0x10]
  mulx  r15, r14, [reg_p2+0x18]
  mulx  rbx, rax, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x8]
  mulx  r9, r8, [reg_p2]
  add   r10, r8
  adc   r11, r9
  mov   [rcx+0x10], r10
  mov   [rcx+0x18], r11  /* c1 contains a0*b1+a1*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   r12, r8
  adc   r13, r9
  mulx  r11, r10, [reg_p2+0x10]
  add   r14, r10
  adc   r15, r11
  mulx  r9, r8, [reg_p2+0x18]
  add   rax, r8
  adc   rbx, r9
  mulx  r11, r10, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x10]
  mulx  r9, r8, [reg_p2]
  add   r12, r8
  adc   r13, r9
  mov   [rcx+0x20], r12
  mov   [rcx+0x28], r13  /* c2 contains a0*b2+a1*b1+a2*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   r14, r8
  adc   r15, r9
  mulx  r13, r12, [reg_p2+0x10]
  add   rax, r12
  adc   rbx, r13
  mulx  r9, r8, [reg_p2+0x18]
  add   r10, r8
  adc   r11, r9
  mulx  r13, r12, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x18]
  mulx  r9, r8, [reg_p2]
  add   r14, r8
  adc   r15, r9
  mov   [rcx+0x30], r14
  mov   [rcx+0x38], r15  /* c3 contains a0*b3+a1*b2+a2*b1+a3*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   rax, r8
  adc   rbx, r9
  mulx  r15, r14, [reg_p2+0x10]
  add   r10, r14
  adc   r11, r15
  mulx  r9, r8, [reg_p2+0x18]
  add   r12, r8
  adc   r13, r9
  mulx  r15, r14, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x20]
  mulx  r9, r8, [reg_p2]
  add   rax, r8
  adc   rbx, r9
  mov   [rcx+0x40], rax
  mov   [rcx+0x48], rbx  /* c4 contains a0*b4+a1*b3+a2*b2+a3*b1+a4*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   r10, r8
  adc   r11, r9
  mulx  rbx, rax, [reg_p2+0x10]
  add   r12, rax
  adc   r13, rbx
  mulx  r9, r8, [reg_p2+0x18]
  add   r14, r8
  adc   r15, r9
  mulx  rbx, rax, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x28]
  mulx  r9, r8, [reg_p2]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1]
  mulx  r9, r8, [reg_p2+0x28]
  add   r10, r8
  adc   r11, r9
  mov   [rcx+0x50], r10
  mov   [rcx+0x58], r11  /* c5 contains a0*b5+a1*b4+...+a4*b1+a5*b0 */
  mulx  r9, r8, [reg_p2+0x30]
  add   r12, r8
  adc   r13, r9
  mulx  r11, r10, [reg_p2+0x38]
  add   r14, r10
  adc   r15, r11
  mulx  r9, r8, [reg_p2+0x40]
  add   rax, r8
  adc   rbx, r9
  mulx  r11, r10, [reg_p2+0x48]

  mov   rdx, [reg_p1+0x8]
  mulx  r9, r8, [reg_p2+0x28]
  add   r12, r8
  adc   r13, r9
  mulx  r9, r8, [reg_p2+0x30]
  add   r14, r8
  adc   r15, r9
  mulx  r9, r8, [reg_p2+0x38]
  add   rax, r8
  adc   rbx, r9
  mulx  r9, r8, [reg_p2+0x40]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1+0x10]
  mulx  r9, r8, [reg_p2+0x28]
  add   r14, r8
  adc   r15, r9
  mulx  r9, r8, [reg_p2+0x30]
  add   rax, r8
  adc   rbx, r9
  mulx  r9, r8, [reg_p2+0x38]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1+0x18]
  mulx  r9, r8, [reg_p2+0x28]
  add   rax, r8
  adc   rbx, r9
  mulx  r9, r8, [reg_p2+0x30]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1+0x20]
  mulx  r9, r8, [reg_p2+0x28]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1+0x28]
  mulx  r9, r8, [reg_p2+0x8]
  add   r12, r8
  adc   r13, r9
  mulx  r9, r8, [reg_p2+0x10]
  add   r14, r8
  adc   r15, r9
  mulx  r9, r8, [reg_p2+0x18]
  add   rax, r8
  adc   rbx, r9
  mulx  r9, r8, [reg_p2+0x20]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1+0x30]
  mulx  r9, r8, [reg_p2]
  add   r12, r8
  adc   r13, r9
  mov   [rcx+0x60], r12
  mov   [rcx+0x68], r13  /* c6 contains a0*b6+a1*b5+...+a5*b1+a6*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   r14, r8
  adc   r15, r9
  mulx  r13, r12, [reg_p2+0x10]
  add   rax, r12
  adc   rbx, r13
  mulx  r9, r8, [reg_p2+0x18]
  add   r10, r8
  adc   r11, r9
  mulx  r13, r12, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x38]
  mulx  r9, r8, [reg_p2]
  add   r14, r8
  adc   r15, r9
  mov   [rcx+0x70], r14
  mov   [rcx+0x78], r15  /* c7 contains a0*b7+a1*b6+...+a6*b1+a7*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   rax, r8
  adc   rbx, r9
  mulx  r15, r14, [reg_p2+0x10]
  add   r10, r14
  adc   r11, r15
  mulx  r9, r8, [reg_p2+0x18]
  add   r12, r8
  adc   r13, r9
  mulx  r15, r14, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x40]
  mulx  r9, r8, [reg_p2]
  add   rax, r8
  adc   rbx, r9
  mov   [rcx+0x80], rax
  mov   [rcx+0x88], rbx  /* c8 contains a0*b8+a1*b7+...+a7*b1+a8*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   r10, r8
  adc   r11, r9
  mulx  rbx, rax, [reg_p2+0x10]
  add   r12, rax
  adc   r13, rbx
  mulx  r9, r8, [reg_p2+0x18]
  add   r14, r8
  adc   r15, r9
  mulx  rbx, rax, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x48]
  mulx  r9, r8, [reg_p2]
  add   r10, r8
  adc   r11, r9
  mov   [rcx+0x90], r10
  mov   [rcx+0x98], r11  /* c9 contains a0*b9+a1*b8+...+a8*b1+a9*b0 */
  mulx  r9, r8, [reg_p2+0x8]
  add   r12, r8
  adc   r13, r9
  mulx  r11, r10, [reg_p2+0x10]
  add   r14, r10
  adc   r15, r11
  mulx  r9, r8, [reg_p2+0x18]
  add   rax, r8
  adc   rbx, r9
  mulx  r11, r10, [reg_p2+0x20]

  mov   rdx, [reg_p1+0x08]
  mulx  r9, r8, [reg_p2+0x48]
  add   r12, r8
  adc   r13, r9

  mov   rdx, [reg_p1+0x10]
  mulx  r9, r8, [reg_p2+0x40]
  add   r12, r8
  adc   r13, r9
  mulx  r9, r8, [reg_p2+0x48]
  add   r14, r8
  adc   r15, r9

  mov   rdx, [reg_p1+0x18]
  mulx  r9, r8, [reg_p2+0x38]
  add   r12, r8
  adc   r13, r9
  mulx  r9, r8, [reg_p2+0x40]
  add   r14, r8
  adc   r15, r9
  mulx  r9, r8, [reg_p2+0x48]
  add   rax, r8
  adc   rbx, r9

  mov   rdx, [reg_p1+0x20]
  mulx  r9, r8, [reg_p2+0x30]
  add   r12, r8
  adc   r13, r9
  mulx  r9, r8, [reg_p2+0x38]
  add   r14, r8
  adc   r15, r9
  mulx  r9, r8, [reg_p2+0x40]
  add   rax, r8
  adc   rbx, r9
  mulx  r9, r8, [reg_p2+0x48]
  add   r10, r8
  adc   r11, r9

  mov   rdx, [reg_p1+0x28]
  mulx  r9, r8, [reg_p2+0x28]
  add   r12, r8
  adc   r13, r9
  mov   r8, [rcx]
  mov   r9, [rcx+0x8]
  add   r8, r12
  adc   r9, r13
  add   r8, r12
  adc   r9, r13
  add   r12, r8
  adc   r13, r9
  mov   [rcx], r12
  mov   [rcx+0x8], r13    /* c0 += 3*(a1*b9+a2*b8+...+a8*b2+a9*b1) */
  mulx  r9, r8, [reg_p2+0x30]
  add   r14, r8
  adc   r15, r9
  mulx  r13, r12, [reg_p2+0x38]
  add   rax, r12
  adc   rbx, r13
  mulx  r9, r8, [reg_p2+0x40]
  add   r10, r8
  adc   r11, r9
  mulx  r13, r12, [reg_p2+0x48]

  mov   rdx, [reg_p1+0x30]
  mulx  r9, r8, [reg_p2+0x28]
  add   r14, r8
  adc   r15, r9
  mov   r8, [rcx+0x10]
  mov   r9, [rcx+0x18]
  add   r8, r14
  adc   r9, r15
  add   r8, r14
  adc   r9, r15
  add   r14, r8
  adc   r15, r9
  mov   [rcx+0x10], r14
  mov   [rcx+0x18], r15   /* c1 += 3*(a2*b9+a3*b8+...+a8*b3+a9*b2) */
  mulx  r9, r8, [reg_p2+0x30]
  add   rax, r8
  adc   rbx, r9
  mulx  r15, r14, [reg_p2+0x38]
  add   r10, r14
  adc   r11, r15
  mulx  r9, r8, [reg_p2+0x40]
  add   r12, r8
  adc   r13, r9
  mulx  r15, r14, [reg_p2+0x48]

  mov   rdx, [reg_p1+0x38]
  mulx  r9, r8, [reg_p2+0x28]
  add   rax, r8
  adc   rbx, r9
  mov   r8, [rcx+0x20]
  mov   r9, [rcx+0x28]
  add   r8, rax
  adc   r9, rbx
  add   r8, rax
  adc   r9, rbx
  add   rax, r8
  adc   rbx, r9
  mov   [rcx+0x20], rax
  mov   [rcx+0x28], rbx   /* c2 += 3*(a3*b9+a4*b8+...+a8*b4+a9*b3) */
  mulx  r9, r8, [reg_p2+0x30]
  add   r10, r8
  adc   r11, r9
  mulx  rbx, rax, [reg_p2+0x38]
  add   r12, rax
  adc   r13, rbx
  mulx  r9, r8, [reg_p2+0x40]
  add   r14, r8
  adc   r15, r9
  mulx  rbx, rax, [reg_p2+0x48]

  mov   rdx, [reg_p1+0x40]
  mulx  r9, r8, [reg_p2+0x28]
  add   r10, r8
  adc   r11, r9
  mov   r8, [rcx+0x30]
  mov   r9, [rcx+0x38]
  add   r8, r10
  adc   r9, r11
  add   r8, r10
  adc   r9, r11
  add   r10, r8
  adc   r11, r9
  mov   [rcx+0x30], r10
  mov   [rcx+0x38], r11   /* c3 += 3*(a4*b9+a5*b8+...+a8*b5+a9*b4) */
  mulx  r9, r8, [reg_p2+0x30]
  add   r12, r8
  adc   r13, r9
  mulx  r11, r10, [reg_p2+0x38]
  add   r14, r10
  adc   r15, r11
  mulx  r9, r8, [reg_p2+0x40]
  add   rax, r8
  adc   rbx, r9
  mulx  r11, r10, [reg_p2+0x48]

  mov   rdx, [reg_p1+0x48]
  mulx  r9, r8, [reg_p2+0x28]
  add   r12, r8
  adc   r13, r9
  mov   r8, [rcx+0x40]
  mov   r9, [rcx+0x48]
  add   r8, r12
  adc   r9, r13
  add   r8, r12
  adc   r9, r13
  add   r12, r8
  adc   r13, r9
  mov   [rcx+0x40], r12
  mov   [rcx+0x48], r13   /* c4 += 3*(a5*b9+a6*b8+a7*b7+a8*b6+a9*b5) */
  mulx  r9, r8, [reg_p2+0x30]
  mulx  r13, r12, [reg_p2+0x38]
  add   r14, r8
  adc   r15, r9
  add   rax, r12
  adc   rbx, r13
  mov   r8, [rcx+0x50]
  mov   r9, [rcx+0x58]
  mov   r12, [rcx+0x60]
  mov   r13, [rcx+0x68]
  add   r8, r14
  adc   r9, r15
  add   r12, rax
  adc   r13, rbx
  add   r8, r14
  adc   r9, r15
  add   r12, rax
  adc   r13, rbx
  add   r14, r8
  adc   r15, r9
  add   rax, r12
  adc   rbx, r13
  mov   [rcx+0x50], r14
  mov   [rcx+0x58], r15   /* c5 += 3*(a6*b9+a7*b8+a8*b7+a9*b6) */
  mov   [rcx+0x60], rax
  mov   [rcx+0x68], rbx   /* c6 += 3*(a7*b9+a8*b8+a9*b7) */

  mulx  r9, r8, [reg_p2+0x40]
  mulx  r13, r12, [reg_p2+0x48]
  add   r10, r8
  adc   r11, r9
  mov   r8, [rcx+0x70]
  mov   r9, [rcx+0x78]
  mov   r14, [rcx+0x80]
  mov   r15, [rcx+0x88]
  add   r8, r10
  adc   r9, r11
  add   r14, r12
  adc   r15, r13
  add   r8, r10
  adc   r9, r11
  add   r14, r12
  adc   r15, r13
  add   r10, r8
  adc   r11, r9
  add   r12, r14
  adc   r13, r15
  mov   [rcx+0x70], r10
  mov   [rcx+0x78], r11   /* c7 += 3*(a8*b9+a9*b8) */
  mov   [rcx+0x80], r12
  mov   [rcx+0x88], r13   /* c8 += 3*(a9*b9) */

  pop   rbx
  pop   r15
  pop   r14
  pop   r13
  pop   r12
  ret
#endif

/*
 * Addition in PMNS without coefficients reduction
 * Operation: c [reg_p3] = a [reg_p1] + b [reg_p2]
 */
/*
.global fmt(mp_add)
fmt(mp_add):

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+8]
  mov   r10, [reg_p1+16]
  mov   r11, [reg_p1+24]
  mov   rax, [reg_p1+32]

  add   r8, [reg_p2]
  add   r9, [reg_p2+8]
  add   r10, [reg_p2+16]
  add   r11, [reg_p2+24]
  add   rax, [reg_p2+32]

  mov   [reg_p3], r8
  mov   [reg_p3+8], r9
  mov   [reg_p3+16], r10
  mov   [reg_p3+24], r11
  mov   [reg_p3+32], rax

  mov   r8, [reg_p1+40]
  mov   r9, [reg_p1+48]
  mov   r10, [reg_p1+66]
  mov   r11, [reg_p1+64]
  mov   rax, [reg_p1+72]

  add   r8, [reg_p2+40]
  add   r9, [reg_p2+48]
  add   r10, [reg_p2+56]
  add   r11, [reg_p2+64]
  add   rax, [reg_p2+72]

  mov   [reg_p3+40], r8
  mov   [reg_p3+48], r9
  mov   [reg_p3+56], r10
  mov   [reg_p3+64], r11
  mov   [reg_p3+72], rax

  ret
*/
