/*
 * SIDH: an efficient supersingular isogeny cryptography library
 * x64 assembly code for P503 using PMNS
 */

.intel_syntax noprefix

/* Format function and variable names for Mac OS X */
#if defined(__APPLE__)
  #define fmt(f) _##f
#else
  #define fmt(f) f
#endif

/* Registers that are used for parameter passing: */
#define reg_p1 rdi
#define reg_p2 rsi
#define reg_p3 rdx

#include "pmns_internal.h"

.text
/*
 * Addition in GF(p) in PMNS
 * Operation: c [reg_p2] = a [reg_p1] + b [reg_p2]
 */
.global fmt(fpadd503)
fmt(fpadd503):
  push  rbx
  push  r12

  mov   rax, (PMNS_K_last_digit_mask >> 1)
  lea   rbx, [rip+fmt(pmns_corrections2)]
  lea   rcx, [rip+fmt(pmns_carries)]

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+0x8]
  mov   r10, [reg_p1+0x10]

  add   r8, [reg_p2]
  adc   r9, [reg_p2+0x8]
  adc   r10, [reg_p2+0x10]
  mov   r11, r10
  and   r10, rax
  shr   r11, PMNS_K_last_digit_shift-1
  shl   r11, 1
  add   r9, [rbx+8*r11]
  adc   r10, [rbx+8*r11+0x8]

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10

  mov   r8, [reg_p1+0x18]
  mov   r9, [reg_p1+0x20]
  mov   r10, [reg_p1+0x28]

  add   r8, [reg_p2+0x18]
  adc   r9, [reg_p2+0x20]
  adc   r10, [reg_p2+0x28]
  mov   r12, r10
  and   r10, rax
  shr   r12, PMNS_K_last_digit_shift-1
  shl   r12, 1
  add   r8, [rcx+4*r11]
  adc   r9, [rbx+8*r12]
  adc   r10, [rbx+8*r12+0x8]

  mov   [reg_p3+0x18], r8
  mov   [reg_p3+0x20], r9
  mov   [reg_p3+0x28], r10

  mov   r8, [reg_p1+0x30]
  mov   r9, [reg_p1+0x38]
  mov   r10, [reg_p1+0x40]

  add   r8, [reg_p2+0x30]
  adc   r9, [reg_p2+0x38]
  adc   r10, [reg_p2+0x40]
  mov   r11, r10
  and   r10, rax
  shr   r11, PMNS_K_last_digit_shift-1
  shl   r11, 1
  add   r8, [rcx+4*r12]
  adc   r9, [rbx+8*r11]
  adc   r10, [rbx+8*r11+0x8]

  mov   [reg_p3+0x30], r8
  mov   [reg_p3+0x38], r9
  mov   [reg_p3+0x40], r10

  mov   r8, [reg_p3]
  mov   r9, [reg_p3+0x8]
  mov   r10, [reg_p3+0x10]
  mov   r12, [rcx+4*r11]
  shl   r12, 2
  add   r8, r12
  adc   r9, 0
  adc   r10, 0

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10

  pop   r12
  pop   rbx

  ret

/*
 * Subtraction in GF(p) in PMNS
 * Operation: c [reg_p2] = a [reg_p1] - b [reg_p2]
 */
.global fmt(fpsub503)
fmt(fpsub503):
  push  rbx
  push  r12
  push  r13
  push  r14
  push  r15

  mov   rax, (PMNS_K_last_digit_mask >> 1)
  lea   rbx, [rip+fmt(pmns_corrections2)]
  lea   rcx, [rip+fmt(pmns_carries)]

  mov r13, 0xfffffffffffffff0
  mov r14, 0x82c029c18cbfffff
  mov r15, 0x0000040223f699c7

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+0x8]
  mov   r10, [reg_p1+0x10]

  add   r8, r13
  adc   r9, r14
  adc   r10, r15
  sub   r8, [reg_p2]
  sbb   r9, [reg_p2+0x8]
  sbb   r10, [reg_p2+0x10]
  mov   r11, r10
  and   r10, rax
  shr   r11, PMNS_K_last_digit_shift-1
  shl   r11, 1
  add   r9, [rbx+8*r11]
  adc   r10, [rbx+8*r11+0x8]

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10

  mov r13, 0xfffffffffffffffc

  mov   r8, [reg_p1+0x18]
  mov   r9, [reg_p1+0x20]
  mov   r10, [reg_p1+0x28]

  add   r8, r13
  adc   r9, r14
  adc   r10, r15
  sub   r8, [reg_p2+0x18]
  sbb   r9, [reg_p2+0x20]
  sbb   r10, [reg_p2+0x28]
  mov   r12, r10
  and   r10, rax
  shr   r12, PMNS_K_last_digit_shift-1
  shl   r12, 1
  add   r8, [rcx+4*r11]
  add   r9, [rbx+8*r12]
  adc   r10, [rbx+8*r12+0x8]

  mov   [reg_p3+0x18], r8
  mov   [reg_p3+0x20], r9
  mov   [reg_p3+0x28], r10

  mov   r8, [reg_p1+0x30]
  mov   r9, [reg_p1+0x38]
  mov   r10, [reg_p1+0x40]

  add   r8, r13
  adc   r9, r14
  adc   r10, r15
  sub   r8, [reg_p2+0x30]
  sbb   r9, [reg_p2+0x38]
  sbb   r10, [reg_p2+0x40]
  mov   r11, r10
  and   r10, rax
  shr   r11, PMNS_K_last_digit_shift-1
  shl   r11, 1
  add   r8, [rcx+4*r12]
  add   r9, [rbx+8*r11]
  adc   r10, [rbx+8*r11+0x8]

  mov   [reg_p3+0x30], r8
  mov   [reg_p3+0x38], r9
  mov   [reg_p3+0x40], r10

  mov   r8, [reg_p3]
  mov   r9, [reg_p3+0x8]
  mov   r10, [reg_p3+0x10]
  mov   r12, [rcx+4*r11]
  shl   r12, 2
  add   r8, r12
  adc   r9, 0
  adc   r10, 0

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10


  pop  r15
  pop  r14
  pop  r13
  pop  r12
  pop  rbx

  ret


/*
 * Division by 2 in GF(p) in PMNS
 * Operation: c [reg_p2] = a [reg_p1] / 2
 */
.global fmt(fpdiv2_503)
fmt(fpdiv2_503):
  push  r12
  push  r13

  mov   r8, 0xe0b00a7063300000
  mov   r9, 0x0000010088fda671

  mov   r12, [reg_p1+0x30]
  mov   r13, [reg_p1+0x38]
  mov   rdx, [reg_p1+0x40]
  shrd  r12, r13, 1
  sbb   rax, rax
  shrd  r13, rdx, 1
  shr   rdx, 1
  mov   [reg_p2+0x30], r12

  mov   r10, [reg_p1+0x18]
  mov   r11, [reg_p1+0x20]
  mov   r12, [reg_p1+0x28]
  mov   rcx, rax
  and   rax, r8
  and   rcx, r9
  add   r11, rax
  adc   r12, rcx
  shrd  r10, r11, 1
  sbb   rax, rax
  shrd  r11, r12, 1
  shr   r12, 1
  mov   [reg_p2+0x18], r10
  mov   [reg_p2+0x20], r11
  mov   [reg_p2+0x28], r12

  mov   r10, [reg_p1]
  mov   r11, [reg_p1+0x08]
  mov   r12, [reg_p1+0x10]
  mov   rcx, rax
  and   rax, r8
  and   rcx, r9
  add   r11, rax
  adc   r12, rcx
  shrd  r10, r11, 1
  sbb   rax, rax
  shrd  r11, r12, 1
  shr   r12, 1
  mov   [reg_p2], r10
  mov   [reg_p2+0x08], r11
  mov   [reg_p2+0x10], r12

  mov   r8, 0x3c16014e0c660000
  mov   r9, 0x00000020111fb4ce
  and   r8, rax
  and   r9, rax
  add   r13, r8
  adc   rdx, r9
  mov   [reg_p2+0x38], r13
  mov   [reg_p2+0x40], rdx

  pop   r13
  pop   r12
  ret


/*
 * Schoolbook multiplication for integers of 192 bits.
 * rdx is reserved for internal usage, registers parameters must not be rdx.
 * rax is reserved for internal usage, registers parameters must not be rax.
 * The result is written in R0, R1, R2, R3, R4 and rax.
 */

.macro MUL192_SCHOOL M0, M1, R0, R1, R2, R3, R4, T0, T1, T2, T3
  mov    rdx, \M0
  mulx   \R1, \R0, \M1      /* R1:R0 = A0*B0 */     /* R0 final */
  mulx   \R3, \T0, 8\M1     /* R3:T0 = A0*B1 */
  xor    rax, rax
  adox   \R1, \T0
  mulx   \T0, \R4, 16\M1    /* T0:R4 = A0*B2 */
  adox   \R3, \R4

  mov    rdx, 8\M0
  mulx   \R4, \T1, \M1      /* R4:T1 = A1*B0 */
  adox   \T0, rax
  xor    rax, rax
  mulx   \T2, \T3, 8\M1     /* T2:T3 = A1*B1 */
  adox   \R1, \T1                                   /* R1 final */
  adcx   \R4, \T3
  mulx   \T3, \R2, 16\M1   /* T3:R2 = A1*B2 */
  adox   \R4, \R3
  adcx   \T2, \R2
  adcx   \T3, rax
  adox   \T2, \T0

  mov    rdx, 16\M0
  mulx   \R3, \R2, \M1      /* R3:R2 = A2*B0 */
  adox   \T3, rax
  xor    rax, rax
  mulx   \T1, \T0, 8\M1     /* T1:T0 = A2*B1 */
  adox   \R2, \R4                                   /* R2 final */
  adcx   \R3, \T2
  mulx   \T2, \R4, 16\M1    /* T2:R4 = A2*B2 */
  adcx   \T1, \T3
  adcx   \T2, rax
  adox   \R3, \T0                                   /* R3 final */
  adox   \R4, \T1                                   /* R4 final */
  adox   rax, \T2                                   /* R5 (=rax) final */
.endm

/*
 * Input is in registers T0, T1, T2 and in immediate constants, I0, I1
 * Output is written in T3, T4, T5, T6, T7
 * All registers must be different.
 * Input registers may be overwritten. 
 * rdx is reserved for internal usage, registers parameters must not be rdx.
 */
.macro MUL192x128imm_SCHOOL R0, R1, R2, T0, T1, T2, T3, T4, Tt, I0, I1
  mov    rdx, \I0
  mulx   \T1, \T0, \R0     /* T1:T0 = I0*R0 */
  mulx   \T2, \Tt, \R1     /* T2:Tt = I0*R1 */
  xor    rax, rax
  adox   \T1, \Tt
  mulx   \T3, \Tt, \R2     /* T3:Tt = I0*R2 */
  adox   \T2, \Tt

  mov    rdx, \I1
  mulx   \R0, \T4, \R0     /* R0:T4 = I1*R0 */
  adox   \T3, rax
  xor    rax, rax
  mulx   \R1, \Tt, \R1     /* R1:Tt = I1*R1 */
  adox   \T1, \T4
  adcx   \R0, \Tt
  mulx   \T4, \R2, \R2     /* T4:R2 = I1*R2 */
  adox   \T2, \R0
  adcx   \R1, \R2
  adcx   \T4, rax
  adox   \T3, \R1
  adox   \T4, rax
.endm

/*
 * Montgomery reduction in PMNS
 * Operation: c [reg_p2] = redc(a [reg_p1])
 */
.global fmt(rdc_mont)
fmt(rdc_mont):
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx

  mov   r8, [reg_p1+96]
  mov   r9, [reg_p1+104]
  mov   r10, [reg_p1+112]

  mov   r11, [reg_p1]
  mov   r12, [reg_p1+8]

  mov   rdx, 0x782C029C18CC0000
  mulx  r14, r13, r11
  add   r9, r13
  adc   r10, r14

  mulx  r13, r12, r12
  add   r10, r12

  mov   rdx, 0x00000040223F699C
  mulx  r13, r12, r11
  add   r10, r12

  mov   r11, [reg_p1+48]
  mov   rdx, 0xAC1B8A4000000000
  mulx  r13, r12, r11
  add   r10, r12

  MUL192x128imm_SCHOOL r8, r9, r10, r11, r12, r13, r14, r15, rbx, 0xE0B00A7063300000, 0x0000010088FDA671

  mov   r10, [reg_p1+48]
  add   r11, [reg_p1+56]
  adc   r12, [reg_p1+64]
  adc   r13, [reg_p1+72]
  adc   r14, [reg_p1+80]
  adc   r15, [reg_p1+88]

  mov   [reg_p2+24], r13
  mov   [reg_p2+32], r14
  mov   [reg_p2+40], r15

  MUL192x128imm_SCHOOL r10, r11, r12, r8, r9, r13, r14, r15, rbx, 0xE0B00A7063300000, 0x0000010088FDA671

  mov   rbx, [reg_p1]
  add   r8, [reg_p1+8]
  adc   r9, [reg_p1+16]
  adc   r13, [reg_p1+24]
  adc   r14, [reg_p1+32]
  adc   r15, [reg_p1+40]

  mov   [reg_p2], r13
  mov   [reg_p2+8], r14
  mov   [reg_p2+16], r15

  MUL192x128imm_SCHOOL rbx, r8, r9, r10, r11, r12, r13, r14, r15, 0x782C029C18CC0000, 0x00000040223F699C

  add   r10, [reg_p1+104]
  adc   r11, [reg_p1+112]
  adc   r12, [reg_p1+120]
  adc   r13, [reg_p1+128]
  adc   r14, [reg_p1+136]

  mov   [reg_p2+48], r12
  mov   [reg_p2+56], r13
  mov   [reg_p2+64], r14

  pop  rbx
  pop  r15
  pop  r14
  pop  r13
  pop  r12

  ret

//#define USE_OPTIMAL_POLYMUL

#ifndef USE_OPTIMAL_POLYMUL
/*
 * Multiplication in GF(p) in PMNS
 * Operation: c [reg_p3] = a [reg_p1] * b [reg_p2]
 */
.global fmt(mp_mul)
fmt(mp_mul):
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx

  mov   rcx, reg_p3

  MUL192_SCHOOL [reg_p1], [reg_p2], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  mov   [rcx], r8
  mov   [rcx+8], r9
  mov   [rcx+16], r10
  mov   [rcx+24], r11
  mov   [rcx+32], r12
  mov   [rcx+40], rax

  MUL192_SCHOOL [reg_p1+24], [reg_p2], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  mov   [rcx+48], r8
  mov   [rcx+56], r9
  mov   [rcx+64], r10
  mov   [rcx+72], r11
  mov   [rcx+80], r12
  mov   [rcx+88], rax

  MUL192_SCHOOL [reg_p1+48], [reg_p2], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  mov   [rcx+96], r8
  mov   [rcx+104], r9
  mov   [rcx+112], r10
  mov   [rcx+120], r11
  mov   [rcx+128], r12
  mov   [rcx+136], rax

  MUL192_SCHOOL [reg_p1], [reg_p2+24], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  add   [rcx+48], r8
  adc   [rcx+56], r9
  adc   [rcx+64], r10
  adc   [rcx+72], r11
  adc   [rcx+80], r12
  adc   [rcx+88], rax

  MUL192_SCHOOL [reg_p1+24], [reg_p2+24], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  add   [rcx+96], r8
  adc   [rcx+104], r9
  adc   [rcx+112], r10
  adc   [rcx+120], r11
  adc   [rcx+128], r12
  adc   [rcx+136], rax


  MUL192_SCHOOL [reg_p1+48], [reg_p2+24], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  shld  rax, r12, 2
  shld  r12, r11, 2
  shld  r11, r10, 2
  shld  r10, r9, 2
  shld  r9, r8, 2
  shl   r8, 2

  add   [rcx], r8
  adc   [rcx+8], r9
  adc   [rcx+16], r10
  adc   [rcx+24], r11
  adc   [rcx+32], r12
  adc   [rcx+40], rax

  MUL192_SCHOOL [reg_p1], [reg_p2+48], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  add   [rcx+96], r8
  adc   [rcx+104], r9
  adc   [rcx+112], r10
  adc   [rcx+120], r11
  adc   [rcx+128], r12
  adc   [rcx+136], rax

  MUL192_SCHOOL [reg_p1+24], [reg_p2+48], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  shld  rax, r12, 2
  shld  r12, r11, 2
  shld  r11, r10, 2
  shld  r10, r9, 2
  shld  r9, r8, 2
  shl   r8, 2

  add   [rcx], r8
  adc   [rcx+8], r9
  adc   [rcx+16], r10
  adc   [rcx+24], r11
  adc   [rcx+32], r12
  adc   [rcx+40], rax


  MUL192_SCHOOL [reg_p1+48], [reg_p2+48], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  shld  rax, r12, 2
  shld  r12, r11, 2
  shld  r11, r10, 2
  shld  r10, r9, 2
  shld  r9, r8, 2
  shl   r8, 2

  add   [rcx+48], r8
  adc   [rcx+56], r9
  adc   [rcx+64], r10
  adc   [rcx+72], r11
  adc   [rcx+80], r12
  adc   [rcx+88], rax

  pop  rbx
  pop  r15
  pop  r14
  pop  r13
  pop  r12

  ret
#else
/*
 * Multiplication in GF(p) in PMNS
 * Operation: c [reg_p3] = a [reg_p1] * b [reg_p2]
 */
.global fmt(mp_mul)
fmt(mp_mul):
  push  r12
  push  r13
  push  r14
  push  r15
  push  rbx

  mov   rcx, reg_p3
  sub   rsp, 0x90

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+0x8]
  mov   r10, [reg_p1+0x10]

  mov   r11, [reg_p1+0x18]
  mov   r12, [reg_p1+0x20]
  mov   r13, [reg_p1+0x28]

  add   r11, r8
  adc   r12, r9
  adc   r13, r10

  /* rsp holds a0+a1 on 3 words */
  mov   [rsp], r11
  mov   [rsp+0x8], r12
  mov   [rsp+0x10], r13

  mov   r11, [reg_p1+0x30]
  mov   r12, [reg_p1+0x38]
  mov   r13, [reg_p1+0x40]

  add   r8, r11
  adc   r9, r12
  adc   r10, r13

  /* rsp+0x18 holds a0+a2 on 3 words */
  mov   [rsp+0x18], r8
  mov   [rsp+0x20], r9
  mov   [rsp+0x28], r10

  add   r11, [reg_p1+0x18]
  adc   r12, [reg_p1+0x20]
  adc   r13, [reg_p1+0x28]

  /* rsp+0x30 holds a1+a2 on 3 words */
  mov   [rsp+0x30], r11
  mov   [rsp+0x38], r12
  mov   [rsp+0x40], r13

  mov   r8, [reg_p2]
  mov   r9, [reg_p2+0x8]
  mov   r10, [reg_p2+0x10]

  mov   r11, [reg_p2+0x18]
  mov   r12, [reg_p2+0x20]
  mov   r13, [reg_p2+0x28]

  add   r11, r8
  adc   r12, r9
  adc   r13, r10

  /* rsp+0x48 holds b0+b1 on 3 words */
  mov   [rsp+0x48], r11
  mov   [rsp+0x50], r12
  mov   [rsp+0x58], r13

  mov   r11, [reg_p2+0x30]
  mov   r12, [reg_p2+0x38]
  mov   r13, [reg_p2+0x40]

  add   r8, r11
  adc   r9, r12
  adc   r10, r13

  /* rsp+0x60 holds b0+b2 on 3 words */
  mov   [rsp+0x60], r8
  mov   [rsp+0x68], r9
  mov   [rsp+0x70], r10

  add   r11, [reg_p2+0x18]
  adc   r12, [reg_p2+0x20]
  adc   r13, [reg_p2+0x28]

  /* rsp+0x78 holds b1+b2 on 3 words */
  mov   [rsp+0x78], r11
  mov   [rsp+0x80], r12
  mov   [rsp+0x88], r13

  MUL192_SCHOOL [rsp], [rsp+0x48], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  mov   [rcx+0x30], r8
  mov   [rcx+0x38], r9
  mov   [rcx+0x40], r10
  mov   [rcx+0x48], r11
  mov   [rcx+0x50], r12
  mov   [rcx+0x58], rax

  MUL192_SCHOOL [rsp+0x18], [rsp+0x60], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  mov   [rcx+0x60], r8
  mov   [rcx+0x68], r9
  mov   [rcx+0x70], r10
  mov   [rcx+0x78], r11
  mov   [rcx+0x80], r12
  mov   [rcx+0x88], rax

  MUL192_SCHOOL [rsp+0x30], [rsp+0x78], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  shld  rax, r12, 2
  shld  r12, r11, 2
  shld  r11, r10, 2
  shld  r10, r9, 2
  shld  r9, r8, 2
  shl   r8, 2

  mov   [rcx], r8
  mov   [rcx+0x8], r9
  mov   [rcx+0x10], r10
  mov   [rcx+0x18], r11
  mov   [rcx+0x20], r12
  mov   [rcx+0x28], rax

  MUL192_SCHOOL [reg_p1], [reg_p2], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  add   [rcx], r8
  adc   [rcx+0x8], r9
  adc   [rcx+0x10], r10
  adc   [rcx+0x18], r11
  adc   [rcx+0x20], r12
  adc   [rcx+0x28], rax

  sub   [rcx+0x30], r8
  sbb   [rcx+0x38], r9
  sbb   [rcx+0x40], r10
  sbb   [rcx+0x48], r11
  sbb   [rcx+0x50], r12
  sbb   [rcx+0x58], rax

  sub   [rcx+0x60], r8
  sbb   [rcx+0x68], r9
  sbb   [rcx+0x70], r10
  sbb   [rcx+0x78], r11
  sbb   [rcx+0x80], r12
  sbb   [rcx+0x88], rax

  MUL192_SCHOOL [reg_p1+0x18], [reg_p2+0x18], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  sub   [rcx+0x30], r8
  sbb   [rcx+0x38], r9
  sbb   [rcx+0x40], r10
  sbb   [rcx+0x48], r11
  sbb   [rcx+0x50], r12
  sbb   [rcx+0x58], rax

  add   [rcx+0x60], r8
  adc   [rcx+0x68], r9
  adc   [rcx+0x70], r10
  adc   [rcx+0x78], r11
  adc   [rcx+0x80], r12
  adc   [rcx+0x88], rax

  shld  rax, r12, 2
  shld  r12, r11, 2
  shld  r11, r10, 2
  shld  r10, r9, 2
  shld  r9, r8, 2
  shl   r8, 2

  sub   [rcx], r8
  sbb   [rcx+0x8], r9
  sbb   [rcx+0x10], r10
  sbb   [rcx+0x18], r11
  sbb   [rcx+0x20], r12
  sbb   [rcx+0x28], rax

  MUL192_SCHOOL [reg_p1+0x30], [reg_p2+0x30], r8, r9, r10, r11, r12, r13, r14, r15, rbx

  sub   [rcx+0x60], r8
  sbb   [rcx+0x68], r9
  sbb   [rcx+0x70], r10
  sbb   [rcx+0x78], r11
  sbb   [rcx+0x80], r12
  sbb   [rcx+0x88], rax

  shld  rax, r12, 2
  shld  r12, r11, 2
  shld  r11, r10, 2
  shld  r10, r9, 2
  shld  r9, r8, 2
  shl   r8, 2

  sub   [rcx], r8
  sbb   [rcx+0x8], r9
  sbb   [rcx+0x10], r10
  sbb   [rcx+0x18], r11
  sbb   [rcx+0x20], r12
  sbb   [rcx+0x28], rax

  add   [rcx+0x30], r8
  adc   [rcx+0x38], r9
  adc   [rcx+0x40], r10
  adc   [rcx+0x48], r11
  adc   [rcx+0x50], r12
  adc   [rcx+0x58], rax

  add   rsp, 0x90

  pop  rbx
  pop  r15
  pop  r14
  pop  r13
  pop  r12

  ret


#endif

/*
 * Double subtraction in PMNS without coefficients reduction
 * Operation: c [reg_p3] = c [reg_p3] - a [reg_p1] - b [reg_p2]
 */
.global fmt(mp_dblsub503x2_pmns)
fmt(mp_dblsub503x2_pmns):
  push  r12

  mov   r8, [reg_p3]
  mov   r9, [reg_p3+0x8]
  mov   r10, [reg_p3+0x10]
  mov   r11, [reg_p3+0x18]
  mov   r12, [reg_p3+0x20]
  mov   rcx, [reg_p3+0x28]

  sub   r8, [reg_p1]
  sbb   r9, [reg_p1+0x8]
  sbb   r10, [reg_p1+0x10]
  sbb   r11, [reg_p1+0x18]
  sbb   r12, [reg_p1+0x20]
  sbb   rcx, [reg_p1+0x28]

  sub   r8, [reg_p2]
  sbb   r9, [reg_p2+0x8]
  sbb   r10, [reg_p2+0x10]
  sbb   r11, [reg_p2+0x18]
  sbb   r12, [reg_p2+0x20]
  sbb   rcx, [reg_p2+0x28]

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10
  mov   [reg_p3+0x18], r11
  mov   [reg_p3+0x20], r12
  mov   [reg_p3+0x28], rcx

  mov   r8, [reg_p3+0x30]
  mov   r9, [reg_p3+0x38]
  mov   r10, [reg_p3+0x40]
  mov   r11, [reg_p3+0x48]
  mov   r12, [reg_p3+0x50]
  mov   rcx, [reg_p3+0x58]

  sub   r8, [reg_p1+0x30]
  sbb   r9, [reg_p1+0x38]
  sbb   r10, [reg_p1+0x40]
  sbb   r11, [reg_p1+0x48]
  sbb   r12, [reg_p1+0x50]
  sbb   rcx, [reg_p1+0x58]

  sub   r8, [reg_p2+0x30]
  sbb   r9, [reg_p2+0x38]
  sbb   r10, [reg_p2+0x40]
  sbb   r11, [reg_p2+0x48]
  sbb   r12, [reg_p2+0x50]
  sbb   rcx, [reg_p2+0x58]

  mov   [reg_p3+0x30], r8
  mov   [reg_p3+0x38], r9
  mov   [reg_p3+0x40], r10
  mov   [reg_p3+0x48], r11
  mov   [reg_p3+0x50], r12
  mov   [reg_p3+0x58], rcx

  mov   r8, [reg_p3+0x60]
  mov   r9, [reg_p3+0x68]
  mov   r10, [reg_p3+0x70]
  mov   r11, [reg_p3+0x78]
  mov   r12, [reg_p3+0x80]
  mov   rcx, [reg_p3+0x88]

  sub   r8, [reg_p1+0x60]
  sbb   r9, [reg_p1+0x68]
  sbb   r10, [reg_p1+0x70]
  sbb   r11, [reg_p1+0x78]
  sbb   r12, [reg_p1+0x80]
  sbb   rcx, [reg_p1+0x88]

  sub   r8, [reg_p2+0x60]
  sbb   r9, [reg_p2+0x68]
  sbb   r10, [reg_p2+0x70]
  sbb   r11, [reg_p2+0x78]
  sbb   r12, [reg_p2+0x80]
  sbb   rcx, [reg_p2+0x88]

  mov   [reg_p3+0x60], r8
  mov   [reg_p3+0x68], r9
  mov   [reg_p3+0x70], r10
  mov   [reg_p3+0x78], r11
  mov   [reg_p3+0x80], r12
  mov   [reg_p3+0x88], rcx

  pop   r12

  ret


/*
 * Substraction in PMNS when coeffs are on two words
 * Operation: c [reg_p2] = a [reg_p1] - b [reg_p2]
 */
.global fmt(mp_sub503x2_pmns)
fmt(mp_sub503x2_pmns):
  push  r12

  mov   r8, [reg_p1]
  mov   r9, [reg_p1+0x8]
  mov   r10, [reg_p1+0x10]
  mov   r11, [reg_p1+0x18]
  mov   r12, [reg_p1+0x20]
  mov   rcx, [reg_p1+0x28]

  add   r8, [rip+fmt(pmns_subx2_cst0)]
  adc   r9, [rip+fmt(pmns_subx2_cst0)+0x8]
  adc   r10, [rip+fmt(pmns_subx2_cst0)+0x10]
  adc   r11, [rip+fmt(pmns_subx2_cst0)+0x18]
  adc   r12, [rip+fmt(pmns_subx2_cst0)+0x20]
  adc   rcx, [rip+fmt(pmns_subx2_cst0)+0x28]

  sub   r8, [reg_p2]
  sbb   r9, [reg_p2+0x8]
  sbb   r10, [reg_p2+0x10]
  sbb   r11, [reg_p2+0x18]
  sbb   r12, [reg_p2+0x20]
  sbb   rcx, [reg_p2+0x28]

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10
  mov   [reg_p3+0x18], r11
  mov   [reg_p3+0x20], r12
  mov   [reg_p3+0x28], rcx

  mov   r8, [reg_p1+0x30]
  mov   r9, [reg_p1+0x38]
  mov   r10, [reg_p1+0x40]
  mov   r11, [reg_p1+0x48]
  mov   r12, [reg_p1+0x50]
  mov   rcx, [reg_p1+0x58]

  add   r8, [rip+fmt(pmns_subx2_cst1)]
  adc   r9, [rip+fmt(pmns_subx2_cst1)+0x8]
  adc   r10, [rip+fmt(pmns_subx2_cst1)+0x10]
  adc   r11, [rip+fmt(pmns_subx2_cst1)+0x18]
  adc   r12, [rip+fmt(pmns_subx2_cst1)+0x20]
  adc   rcx, [rip+fmt(pmns_subx2_cst1)+0x28]

  sub   r8, [reg_p2+0x30]
  sbb   r9, [reg_p2+0x38]
  sbb   r10, [reg_p2+0x40]
  sbb   r11, [reg_p2+0x48]
  sbb   r12, [reg_p2+0x50]
  sbb   rcx, [reg_p2+0x58]

  mov   [reg_p3+0x30], r8
  mov   [reg_p3+0x38], r9
  mov   [reg_p3+0x40], r10
  mov   [reg_p3+0x48], r11
  mov   [reg_p3+0x50], r12
  mov   [reg_p3+0x58], rcx

  mov   r8, [reg_p1+0x60]
  mov   r9, [reg_p1+0x68]
  mov   r10, [reg_p1+0x70]
  mov   r11, [reg_p1+0x78]
  mov   r12, [reg_p1+0x80]
  mov   rcx, [reg_p1+0x88]

  add   r8, [rip+fmt(pmns_subx2_cst1)]
  adc   r9, [rip+fmt(pmns_subx2_cst1)+0x8]
  adc   r10, [rip+fmt(pmns_subx2_cst1)+0x10]
  adc   r11, [rip+fmt(pmns_subx2_cst1)+0x18]
  adc   r12, [rip+fmt(pmns_subx2_cst1)+0x20]
  adc   rcx, [rip+fmt(pmns_subx2_cst1)+0x28]

  sub   r8, [reg_p2+0x60]
  sbb   r9, [reg_p2+0x68]
  sbb   r10, [reg_p2+0x70]
  sbb   r11, [reg_p2+0x78]
  sbb   r12, [reg_p2+0x80]
  sbb   rcx, [reg_p2+0x88]

  mov   [reg_p3+0x60], r8
  mov   [reg_p3+0x68], r9
  mov   [reg_p3+0x70], r10
  mov   [reg_p3+0x78], r11
  mov   [reg_p3+0x80], r12
  mov   [reg_p3+0x88], rcx

  pop   r12

  ret

/*
 * Addition in PMNS without coefficients reduction
 * Operation: c [reg_p3] = a [reg_p1] + b [reg_p2]
 */
.global fmt(mp_add)
fmt(mp_add):
  mov   r8, [reg_p1]
  mov   r9, [reg_p1+0x8]
  mov   r10, [reg_p1+0x10]

  add   r8, [reg_p2]
  adc   r9, [reg_p2+0x8]
  adc   r10, [reg_p2+0x10]

  mov   [reg_p3], r8
  mov   [reg_p3+0x8], r9
  mov   [reg_p3+0x10], r10

  mov   r8, [reg_p1+0x18]
  mov   r9, [reg_p1+0x20]
  mov   r10, [reg_p1+0x28]

  add   r8, [reg_p2+0x18]
  adc   r9, [reg_p2+0x20]
  adc   r10, [reg_p2+0x28]

  mov   [reg_p3+0x18], r8
  mov   [reg_p3+0x20], r9
  mov   [reg_p3+0x28], r10

  mov   r8, [reg_p1+0x30]
  mov   r9, [reg_p1+0x38]
  mov   r10, [reg_p1+0x40]

  add   r8, [reg_p2+0x30]
  adc   r9, [reg_p2+0x38]
  adc   r10, [reg_p2+0x40]

  mov   [reg_p3+0x30], r8
  mov   [reg_p3+0x38], r9
  mov   [reg_p3+0x40], r10

  ret
